from time import sleep, time

from ratio1 import Payload, Pipeline, Session, load_dotenv

boxes = {}


def instance_on_data(pipeline: Pipeline, data: Payload):
  global boxes
  data['INSTANCES'].sort()
  boxes[data['BOX']] = (data['INSTANCES'], time())
  return


def pipeline_on_notification(pipeline: Pipeline, notification: dict):
  pipeline.P(
      "Received specific notification for pipeline {}".format(pipeline.name))
  return


env_variables = load_dotenv(load_env=False)
listener_params = {
  'HOST': env_variables['AIXP_HOSTNAME'],
  'PORT': int(env_variables['AIXP_PORT']),
  'USER': env_variables['AIXP_USERNAME'],
  'PASS': env_variables['AIXP_PASSWORD'],
  'TOPIC': 'lummetry/ctrl',
}

node_id = 'node_id'
sess = Session()

pipeline = sess.attach_to_pipeline(
  node=node_id,
  name='test_mqtt',
  on_notification=pipeline_on_notification,
  max_wait_time=60
)

# now start a cyclic process
inst = pipeline.attach_to_custom_plugin_instance('inst01', on_data=instance_on_data)


def generate_net_map():
  global boxes
  net_map_msgs = ["Printing network map"]
  for box, (instances, last_time) in boxes.items():
    if last_time - time() > 30:
      continue
    net_map_msgs.append(box + ":")
    for i in range(len(instances)):
      (pipeline, signature, instance) = instances[i]

      (o_pipeline, o_signature) = (None, None)
      if i != 0:
        (o_pipeline, o_signature, _) = instances[i - 1]

      same_pipeline = pipeline == o_pipeline
      same_signature = signature == o_signature

      if not same_pipeline:
        net_map_msgs.append(" --- " + pipeline + ":")
      if not same_pipeline or not same_signature:
        net_map_msgs.append("    --- " + signature + ":")
      net_map_msgs.append("        --- " + instance)

    net_map_msgs.append("")
  return "\n".join(net_map_msgs)


try:
  while True:
    sleep(10)

    pipeline.P(generate_net_map(), color='m')
except KeyboardInterrupt:
  pipeline.P("CTRL+C detected. Closing example..", color='r')


pipeline.close()
sess.close()



from .base import Payload
from .base import Pipeline
from .base import Instance
from .base import CustomPluginTemplate
from .base import DistributedCustomCodePresets
from .default import MqttSession as Session
from .utils import load_dotenv
from ._ver import __VER__ as version
from ._ver import __VER__ as __version__
from .base_decentra_object import BaseDecentrAIObject
from .plugins_manager_mixin import _PluginsManagerMixin
from .logging import Logger
from .code_cheker import BaseCodeChecker
from .const import PLUGIN_SIGNATURES, PAYLOAD_DATA
from .const import HB as HEARTBEAT_DATA
from .default.instance import PLUGIN_TYPES



import inspect

class _ClassInstanceMixin(object):
  """
  Mixin for class instance functionalities that are attached to `ratio1.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `ratio1.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_ClassInstanceMixin, self).__init__()

  @staticmethod
  def get_class_instance_methods(obj, spacing=20):
    methodList = []
    for method_name in dir(obj):
      try:
        if callable(getattr(obj, method_name)):
          methodList.append(str(method_name))
      except:
        methodList.append(str(method_name))
    processFunc = (lambda s: ' '.join(s.split())) or (lambda s: s)
    for method in methodList:
      try:
        print(str(method.ljust(spacing)) + ' ' + processFunc(str(getattr(obj, method).__doc__)[0:90]))
      except:
        print(method.ljust(spacing) + ' ' + ' getattr() failed')
    return


  @staticmethod
  def get_class_instance_params(obj, n=None):
    """
    Parameters
    ----------
    obj : any type
      the inspected object.
    n : int, optional
      the number of params that are returned. The default is None
      (all params returned).

    Returns
    -------
    out_str : str
      the description of the object 'obj' in terms of parameters values.
    """

    out_str = obj.__class__.__name__ + "("
    n_added_to_log = 0
    for _iter, (prop, value) in enumerate(vars(obj).items()):
      if type(value) in [int, float, bool]:
        out_str += prop + '=' + str(value) + ','
        n_added_to_log += 1
      elif type(value) in [str]:
        out_str += prop + "='" + value + "',"
        n_added_to_log += 1

      if n is not None and n_added_to_log >= n:
        break
    # endfor

    out_str = out_str[:-1] if out_str[-1] == ',' else out_str
    out_str += ')'
    return out_str

  @staticmethod
  def get_object_params(*args, **kwargs):
    print("DeprecationWarning! `get_object_params` is deprecated. Please use `get_class_instance_params` instead")
    return _ClassInstanceMixin.get_class_instance_params(*args, **kwargs)

  @staticmethod
  def get_object_methods(*args, **kwargs):
    print("DeprecationWarning! `get_object_methods` is deprecated. Please use `get_class_instance_methods` instead")
    return _ClassInstanceMixin.get_class_instance_methods(*args, **kwargs)

  @staticmethod
  def get_class_methods(cls, include_parent=True):
    lst_methods = inspect.getmembers(cls, predicate=inspect.isfunction)

    if not include_parent:
      lst_methods = list(filter(
        lambda x: cls.__name__ in x[1].__qualname__,
        lst_methods
      ))

    return lst_methods






# This file is autogenerated by the update_windows_mapping.py script
# Do not edit.
win_tz = {
    "AUS Central Standard Time": "Australia/Darwin",
    "AUS Eastern Standard Time": "Australia/Sydney",
    "Afghanistan Standard Time": "Asia/Kabul",
    "Alaskan Standard Time": "America/Anchorage",
    "Aleutian Standard Time": "America/Adak",
    "Altai Standard Time": "Asia/Barnaul",
    "Arab Standard Time": "Asia/Riyadh",
    "Arabian Standard Time": "Asia/Dubai",
    "Arabic Standard Time": "Asia/Baghdad",
    "Argentina Standard Time": "America/Buenos_Aires",
    "Astrakhan Standard Time": "Europe/Astrakhan",
    "Atlantic Standard Time": "America/Halifax",
    "Aus Central W. Standard Time": "Australia/Eucla",
    "Azerbaijan Standard Time": "Asia/Baku",
    "Azores Standard Time": "Atlantic/Azores",
    "Bahia Standard Time": "America/Bahia",
    "Bangladesh Standard Time": "Asia/Dhaka",
    "Belarus Standard Time": "Europe/Minsk",
    "Bougainville Standard Time": "Pacific/Bougainville",
    "Canada Central Standard Time": "America/Regina",
    "Cape Verde Standard Time": "Atlantic/Cape_Verde",
    "Caucasus Standard Time": "Asia/Yerevan",
    "Cen. Australia Standard Time": "Australia/Adelaide",
    "Central America Standard Time": "America/Guatemala",
    "Central Asia Standard Time": "Asia/Almaty",
    "Central Brazilian Standard Time": "America/Cuiaba",
    "Central Europe Standard Time": "Europe/Budapest",
    "Central European Standard Time": "Europe/Warsaw",
    "Central Pacific Standard Time": "Pacific/Guadalcanal",
    "Central Standard Time": "America/Chicago",
    "Central Standard Time (Mexico)": "America/Mexico_City",
    "Chatham Islands Standard Time": "Pacific/Chatham",
    "China Standard Time": "Asia/Shanghai",
    "Cuba Standard Time": "America/Havana",
    "Dateline Standard Time": "Etc/GMT+12",
    "E. Africa Standard Time": "Africa/Nairobi",
    "E. Australia Standard Time": "Australia/Brisbane",
    "E. Europe Standard Time": "Europe/Chisinau",
    "E. South America Standard Time": "America/Sao_Paulo",
    "Easter Island Standard Time": "Pacific/Easter",
    "Eastern Standard Time": "America/New_York",
    "Eastern Standard Time (Mexico)": "America/Cancun",
    "Egypt Standard Time": "Africa/Cairo",
    "Ekaterinburg Standard Time": "Asia/Yekaterinburg",
    "FLE Standard Time": "Europe/Kiev",
    "Fiji Standard Time": "Pacific/Fiji",
    "GMT Standard Time": "Europe/London",
    "GTB Standard Time": "Europe/Bucharest",
    "Georgian Standard Time": "Asia/Tbilisi",
    "Greenland Standard Time": "America/Godthab",
    "Greenwich Standard Time": "Atlantic/Reykjavik",
    "Haiti Standard Time": "America/Port-au-Prince",
    "Hawaiian Standard Time": "Pacific/Honolulu",
    "India Standard Time": "Asia/Calcutta",
    "Iran Standard Time": "Asia/Tehran",
    "Israel Standard Time": "Asia/Jerusalem",
    "Jordan Standard Time": "Asia/Amman",
    "Kaliningrad Standard Time": "Europe/Kaliningrad",
    "Korea Standard Time": "Asia/Seoul",
    "Libya Standard Time": "Africa/Tripoli",
    "Line Islands Standard Time": "Pacific/Kiritimati",
    "Lord Howe Standard Time": "Australia/Lord_Howe",
    "Magadan Standard Time": "Asia/Magadan",
    "Magallanes Standard Time": "America/Punta_Arenas",
    "Marquesas Standard Time": "Pacific/Marquesas",
    "Mauritius Standard Time": "Indian/Mauritius",
    "Middle East Standard Time": "Asia/Beirut",
    "Montevideo Standard Time": "America/Montevideo",
    "Morocco Standard Time": "Africa/Casablanca",
    "Mountain Standard Time": "America/Denver",
    "Mountain Standard Time (Mexico)": "America/Mazatlan",
    "Myanmar Standard Time": "Asia/Rangoon",
    "N. Central Asia Standard Time": "Asia/Novosibirsk",
    "Namibia Standard Time": "Africa/Windhoek",
    "Nepal Standard Time": "Asia/Katmandu",
    "New Zealand Standard Time": "Pacific/Auckland",
    "Newfoundland Standard Time": "America/St_Johns",
    "Norfolk Standard Time": "Pacific/Norfolk",
    "North Asia East Standard Time": "Asia/Irkutsk",
    "North Asia Standard Time": "Asia/Krasnoyarsk",
    "North Korea Standard Time": "Asia/Pyongyang",
    "Omsk Standard Time": "Asia/Omsk",
    "Pacific SA Standard Time": "America/Santiago",
    "Pacific Standard Time": "America/Los_Angeles",
    "Pacific Standard Time (Mexico)": "America/Tijuana",
    "Pakistan Standard Time": "Asia/Karachi",
    "Paraguay Standard Time": "America/Asuncion",
    "Qyzylorda Standard Time": "Asia/Qyzylorda",
    "Romance Standard Time": "Europe/Paris",
    "Russia Time Zone 10": "Asia/Srednekolymsk",
    "Russia Time Zone 11": "Asia/Kamchatka",
    "Russia Time Zone 3": "Europe/Samara",
    "Russian Standard Time": "Europe/Moscow",
    "SA Eastern Standard Time": "America/Cayenne",
    "SA Pacific Standard Time": "America/Bogota",
    "SA Western Standard Time": "America/La_Paz",
    "SE Asia Standard Time": "Asia/Bangkok",
    "Saint Pierre Standard Time": "America/Miquelon",
    "Sakhalin Standard Time": "Asia/Sakhalin",
    "Samoa Standard Time": "Pacific/Apia",
    "Sao Tome Standard Time": "Africa/Sao_Tome",
    "Saratov Standard Time": "Europe/Saratov",
    "Singapore Standard Time": "Asia/Singapore",
    "South Africa Standard Time": "Africa/Johannesburg",
    "South Sudan Standard Time": "Africa/Juba",
    "Sri Lanka Standard Time": "Asia/Colombo",
    "Sudan Standard Time": "Africa/Khartoum",
    "Syria Standard Time": "Asia/Damascus",
    "Taipei Standard Time": "Asia/Taipei",
    "Tasmania Standard Time": "Australia/Hobart",
    "Tocantins Standard Time": "America/Araguaina",
    "Tokyo Standard Time": "Asia/Tokyo",
    "Tomsk Standard Time": "Asia/Tomsk",
    "Tonga Standard Time": "Pacific/Tongatapu",
    "Transbaikal Standard Time": "Asia/Chita",
    "Turkey Standard Time": "Europe/Istanbul",
    "Turks And Caicos Standard Time": "America/Grand_Turk",
    "US Eastern Standard Time": "America/Indianapolis",
    "US Mountain Standard Time": "America/Phoenix",
    "UTC": "Etc/UTC",
    "UTC+12": "Etc/GMT-12",
    "UTC+13": "Etc/GMT-13",
    "UTC-02": "Etc/GMT+2",
    "UTC-08": "Etc/GMT+8",
    "UTC-09": "Etc/GMT+9",
    "UTC-11": "Etc/GMT+11",
    "Ulaanbaatar Standard Time": "Asia/Ulaanbaatar",
    "Venezuela Standard Time": "America/Caracas",
    "Vladivostok Standard Time": "Asia/Vladivostok",
    "Volgograd Standard Time": "Europe/Volgograd",
    "W. Australia Standard Time": "Australia/Perth",
    "W. Central Africa Standard Time": "Africa/Lagos",
    "W. Europe Standard Time": "Europe/Berlin",
    "W. Mongolia Standard Time": "Asia/Hovd",
    "West Asia Standard Time": "Asia/Tashkent",
    "West Bank Standard Time": "Asia/Hebron",
    "West Pacific Standard Time": "Pacific/Port_Moresby",
    "Yakutsk Standard Time": "Asia/Yakutsk",
    "Yukon Standard Time": "America/Whitehorse",
}

# Old name for the win_tz variable:
tz_names = win_tz

tz_win = {
    "": "Central Standard Time (Mexico)",
    "Africa/Abidjan": "Greenwich Standard Time",
    "Africa/Accra": "Greenwich Standard Time",
    "Africa/Addis_Ababa": "E. Africa Standard Time",
    "Africa/Algiers": "W. Central Africa Standard Time",
    "Africa/Asmara": "E. Africa Standard Time",
    "Africa/Asmera": "E. Africa Standard Time",
    "Africa/Bamako": "Greenwich Standard Time",
    "Africa/Bangui": "W. Central Africa Standard Time",
    "Africa/Banjul": "Greenwich Standard Time",
    "Africa/Bissau": "Greenwich Standard Time",
    "Africa/Blantyre": "South Africa Standard Time",
    "Africa/Brazzaville": "W. Central Africa Standard Time",
    "Africa/Bujumbura": "South Africa Standard Time",
    "Africa/Cairo": "Egypt Standard Time",
    "Africa/Casablanca": "Morocco Standard Time",
    "Africa/Ceuta": "Romance Standard Time",
    "Africa/Conakry": "Greenwich Standard Time",
    "Africa/Dakar": "Greenwich Standard Time",
    "Africa/Dar_es_Salaam": "E. Africa Standard Time",
    "Africa/Djibouti": "E. Africa Standard Time",
    "Africa/Douala": "W. Central Africa Standard Time",
    "Africa/El_Aaiun": "Morocco Standard Time",
    "Africa/Freetown": "Greenwich Standard Time",
    "Africa/Gaborone": "South Africa Standard Time",
    "Africa/Harare": "South Africa Standard Time",
    "Africa/Johannesburg": "South Africa Standard Time",
    "Africa/Juba": "South Sudan Standard Time",
    "Africa/Kampala": "E. Africa Standard Time",
    "Africa/Khartoum": "Sudan Standard Time",
    "Africa/Kigali": "South Africa Standard Time",
    "Africa/Kinshasa": "W. Central Africa Standard Time",
    "Africa/Lagos": "W. Central Africa Standard Time",
    "Africa/Libreville": "W. Central Africa Standard Time",
    "Africa/Lome": "Greenwich Standard Time",
    "Africa/Luanda": "W. Central Africa Standard Time",
    "Africa/Lubumbashi": "South Africa Standard Time",
    "Africa/Lusaka": "South Africa Standard Time",
    "Africa/Malabo": "W. Central Africa Standard Time",
    "Africa/Maputo": "South Africa Standard Time",
    "Africa/Maseru": "South Africa Standard Time",
    "Africa/Mbabane": "South Africa Standard Time",
    "Africa/Mogadishu": "E. Africa Standard Time",
    "Africa/Monrovia": "Greenwich Standard Time",
    "Africa/Nairobi": "E. Africa Standard Time",
    "Africa/Ndjamena": "W. Central Africa Standard Time",
    "Africa/Niamey": "W. Central Africa Standard Time",
    "Africa/Nouakchott": "Greenwich Standard Time",
    "Africa/Ouagadougou": "Greenwich Standard Time",
    "Africa/Porto-Novo": "W. Central Africa Standard Time",
    "Africa/Sao_Tome": "Sao Tome Standard Time",
    "Africa/Timbuktu": "Greenwich Standard Time",
    "Africa/Tripoli": "Libya Standard Time",
    "Africa/Tunis": "W. Central Africa Standard Time",
    "Africa/Windhoek": "Namibia Standard Time",
    "America/Adak": "Aleutian Standard Time",
    "America/Anchorage": "Alaskan Standard Time",
    "America/Anguilla": "SA Western Standard Time",
    "America/Antigua": "SA Western Standard Time",
    "America/Araguaina": "Tocantins Standard Time",
    "America/Argentina/La_Rioja": "Argentina Standard Time",
    "America/Argentina/Rio_Gallegos": "Argentina Standard Time",
    "America/Argentina/Salta": "Argentina Standard Time",
    "America/Argentina/San_Juan": "Argentina Standard Time",
    "America/Argentina/San_Luis": "Argentina Standard Time",
    "America/Argentina/Tucuman": "Argentina Standard Time",
    "America/Argentina/Ushuaia": "Argentina Standard Time",
    "America/Aruba": "SA Western Standard Time",
    "America/Asuncion": "Paraguay Standard Time",
    "America/Atikokan": "SA Pacific Standard Time",
    "America/Atka": "Aleutian Standard Time",
    "America/Bahia": "Bahia Standard Time",
    "America/Bahia_Banderas": "Central Standard Time (Mexico)",
    "America/Barbados": "SA Western Standard Time",
    "America/Belem": "SA Eastern Standard Time",
    "America/Belize": "Central America Standard Time",
    "America/Blanc-Sablon": "SA Western Standard Time",
    "America/Boa_Vista": "SA Western Standard Time",
    "America/Bogota": "SA Pacific Standard Time",
    "America/Boise": "Mountain Standard Time",
    "America/Buenos_Aires": "Argentina Standard Time",
    "America/Cambridge_Bay": "Mountain Standard Time",
    "America/Campo_Grande": "Central Brazilian Standard Time",
    "America/Cancun": "Eastern Standard Time (Mexico)",
    "America/Caracas": "Venezuela Standard Time",
    "America/Catamarca": "Argentina Standard Time",
    "America/Cayenne": "SA Eastern Standard Time",
    "America/Cayman": "SA Pacific Standard Time",
    "America/Chicago": "Central Standard Time",
    "America/Chihuahua": "Central Standard Time (Mexico)",
    "America/Ciudad_Juarez": "Mountain Standard Time",
    "America/Coral_Harbour": "SA Pacific Standard Time",
    "America/Cordoba": "Argentina Standard Time",
    "America/Costa_Rica": "Central America Standard Time",
    "America/Creston": "US Mountain Standard Time",
    "America/Cuiaba": "Central Brazilian Standard Time",
    "America/Curacao": "SA Western Standard Time",
    "America/Danmarkshavn": "Greenwich Standard Time",
    "America/Dawson": "Yukon Standard Time",
    "America/Dawson_Creek": "US Mountain Standard Time",
    "America/Denver": "Mountain Standard Time",
    "America/Detroit": "Eastern Standard Time",
    "America/Dominica": "SA Western Standard Time",
    "America/Edmonton": "Mountain Standard Time",
    "America/Eirunepe": "SA Pacific Standard Time",
    "America/El_Salvador": "Central America Standard Time",
    "America/Ensenada": "Pacific Standard Time (Mexico)",
    "America/Fort_Nelson": "US Mountain Standard Time",
    "America/Fortaleza": "SA Eastern Standard Time",
    "America/Glace_Bay": "Atlantic Standard Time",
    "America/Godthab": "Greenland Standard Time",
    "America/Goose_Bay": "Atlantic Standard Time",
    "America/Grand_Turk": "Turks And Caicos Standard Time",
    "America/Grenada": "SA Western Standard Time",
    "America/Guadeloupe": "SA Western Standard Time",
    "America/Guatemala": "Central America Standard Time",
    "America/Guayaquil": "SA Pacific Standard Time",
    "America/Guyana": "SA Western Standard Time",
    "America/Halifax": "Atlantic Standard Time",
    "America/Havana": "Cuba Standard Time",
    "America/Hermosillo": "US Mountain Standard Time",
    "America/Indiana/Knox": "Central Standard Time",
    "America/Indiana/Marengo": "US Eastern Standard Time",
    "America/Indiana/Petersburg": "Eastern Standard Time",
    "America/Indiana/Tell_City": "Central Standard Time",
    "America/Indiana/Vevay": "US Eastern Standard Time",
    "America/Indiana/Vincennes": "Eastern Standard Time",
    "America/Indiana/Winamac": "Eastern Standard Time",
    "America/Indianapolis": "US Eastern Standard Time",
    "America/Inuvik": "Mountain Standard Time",
    "America/Iqaluit": "Eastern Standard Time",
    "America/Jamaica": "SA Pacific Standard Time",
    "America/Jujuy": "Argentina Standard Time",
    "America/Juneau": "Alaskan Standard Time",
    "America/Kentucky/Monticello": "Eastern Standard Time",
    "America/Knox_IN": "Central Standard Time",
    "America/Kralendijk": "SA Western Standard Time",
    "America/La_Paz": "SA Western Standard Time",
    "America/Lima": "SA Pacific Standard Time",
    "America/Los_Angeles": "Pacific Standard Time",
    "America/Louisville": "Eastern Standard Time",
    "America/Lower_Princes": "SA Western Standard Time",
    "America/Maceio": "SA Eastern Standard Time",
    "America/Managua": "Central America Standard Time",
    "America/Manaus": "SA Western Standard Time",
    "America/Marigot": "SA Western Standard Time",
    "America/Martinique": "SA Western Standard Time",
    "America/Matamoros": "Central Standard Time",
    "America/Mazatlan": "Mountain Standard Time (Mexico)",
    "America/Mendoza": "Argentina Standard Time",
    "America/Menominee": "Central Standard Time",
    "America/Merida": "Central Standard Time (Mexico)",
    "America/Metlakatla": "Alaskan Standard Time",
    "America/Mexico_City": "Central Standard Time (Mexico)",
    "America/Miquelon": "Saint Pierre Standard Time",
    "America/Moncton": "Atlantic Standard Time",
    "America/Monterrey": "Central Standard Time (Mexico)",
    "America/Montevideo": "Montevideo Standard Time",
    "America/Montreal": "Eastern Standard Time",
    "America/Montserrat": "SA Western Standard Time",
    "America/Nassau": "Eastern Standard Time",
    "America/New_York": "Eastern Standard Time",
    "America/Nipigon": "Eastern Standard Time",
    "America/Nome": "Alaskan Standard Time",
    "America/Noronha": "UTC-02",
    "America/North_Dakota/Beulah": "Central Standard Time",
    "America/North_Dakota/Center": "Central Standard Time",
    "America/North_Dakota/New_Salem": "Central Standard Time",
    "America/Ojinaga": "Central Standard Time",
    "America/Panama": "SA Pacific Standard Time",
    "America/Pangnirtung": "Eastern Standard Time",
    "America/Paramaribo": "SA Eastern Standard Time",
    "America/Phoenix": "US Mountain Standard Time",
    "America/Port-au-Prince": "Haiti Standard Time",
    "America/Port_of_Spain": "SA Western Standard Time",
    "America/Porto_Acre": "SA Pacific Standard Time",
    "America/Porto_Velho": "SA Western Standard Time",
    "America/Puerto_Rico": "SA Western Standard Time",
    "America/Punta_Arenas": "Magallanes Standard Time",
    "America/Rainy_River": "Central Standard Time",
    "America/Rankin_Inlet": "Central Standard Time",
    "America/Recife": "SA Eastern Standard Time",
    "America/Regina": "Canada Central Standard Time",
    "America/Resolute": "Central Standard Time",
    "America/Rio_Branco": "SA Pacific Standard Time",
    "America/Santa_Isabel": "Pacific Standard Time (Mexico)",
    "America/Santarem": "SA Eastern Standard Time",
    "America/Santiago": "Pacific SA Standard Time",
    "America/Santo_Domingo": "SA Western Standard Time",
    "America/Sao_Paulo": "E. South America Standard Time",
    "America/Scoresbysund": "Azores Standard Time",
    "America/Shiprock": "Mountain Standard Time",
    "America/Sitka": "Alaskan Standard Time",
    "America/St_Barthelemy": "SA Western Standard Time",
    "America/St_Johns": "Newfoundland Standard Time",
    "America/St_Kitts": "SA Western Standard Time",
    "America/St_Lucia": "SA Western Standard Time",
    "America/St_Thomas": "SA Western Standard Time",
    "America/St_Vincent": "SA Western Standard Time",
    "America/Swift_Current": "Canada Central Standard Time",
    "America/Tegucigalpa": "Central America Standard Time",
    "America/Thule": "Atlantic Standard Time",
    "America/Thunder_Bay": "Eastern Standard Time",
    "America/Tijuana": "Pacific Standard Time (Mexico)",
    "America/Toronto": "Eastern Standard Time",
    "America/Tortola": "SA Western Standard Time",
    "America/Vancouver": "Pacific Standard Time",
    "America/Virgin": "SA Western Standard Time",
    "America/Whitehorse": "Yukon Standard Time",
    "America/Winnipeg": "Central Standard Time",
    "America/Yakutat": "Alaskan Standard Time",
    "America/Yellowknife": "Mountain Standard Time",
    "Antarctica/Casey": "Central Pacific Standard Time",
    "Antarctica/Davis": "SE Asia Standard Time",
    "Antarctica/DumontDUrville": "West Pacific Standard Time",
    "Antarctica/Macquarie": "Tasmania Standard Time",
    "Antarctica/Mawson": "West Asia Standard Time",
    "Antarctica/McMurdo": "New Zealand Standard Time",
    "Antarctica/Palmer": "SA Eastern Standard Time",
    "Antarctica/Rothera": "SA Eastern Standard Time",
    "Antarctica/South_Pole": "New Zealand Standard Time",
    "Antarctica/Syowa": "Arab Standard Time",
    "Antarctica/Vostok": "Central Asia Standard Time",
    "Arctic/Longyearbyen": "W. Europe Standard Time",
    "Asia/Aden": "Arab Standard Time",
    "Asia/Almaty": "Central Asia Standard Time",
    "Asia/Amman": "Jordan Standard Time",
    "Asia/Anadyr": "Russia Time Zone 11",
    "Asia/Aqtau": "West Asia Standard Time",
    "Asia/Aqtobe": "West Asia Standard Time",
    "Asia/Ashgabat": "West Asia Standard Time",
    "Asia/Ashkhabad": "West Asia Standard Time",
    "Asia/Atyrau": "West Asia Standard Time",
    "Asia/Baghdad": "Arabic Standard Time",
    "Asia/Bahrain": "Arab Standard Time",
    "Asia/Baku": "Azerbaijan Standard Time",
    "Asia/Bangkok": "SE Asia Standard Time",
    "Asia/Barnaul": "Altai Standard Time",
    "Asia/Beirut": "Middle East Standard Time",
    "Asia/Bishkek": "Central Asia Standard Time",
    "Asia/Brunei": "Singapore Standard Time",
    "Asia/Calcutta": "India Standard Time",
    "Asia/Chita": "Transbaikal Standard Time",
    "Asia/Choibalsan": "Ulaanbaatar Standard Time",
    "Asia/Chongqing": "China Standard Time",
    "Asia/Chungking": "China Standard Time",
    "Asia/Colombo": "Sri Lanka Standard Time",
    "Asia/Dacca": "Bangladesh Standard Time",
    "Asia/Damascus": "Syria Standard Time",
    "Asia/Dhaka": "Bangladesh Standard Time",
    "Asia/Dili": "Tokyo Standard Time",
    "Asia/Dubai": "Arabian Standard Time",
    "Asia/Dushanbe": "West Asia Standard Time",
    "Asia/Famagusta": "GTB Standard Time",
    "Asia/Gaza": "West Bank Standard Time",
    "Asia/Harbin": "China Standard Time",
    "Asia/Hebron": "West Bank Standard Time",
    "Asia/Hong_Kong": "China Standard Time",
    "Asia/Hovd": "W. Mongolia Standard Time",
    "Asia/Irkutsk": "North Asia East Standard Time",
    "Asia/Istanbul": "Turkey Standard Time",
    "Asia/Jakarta": "SE Asia Standard Time",
    "Asia/Jayapura": "Tokyo Standard Time",
    "Asia/Jerusalem": "Israel Standard Time",
    "Asia/Kabul": "Afghanistan Standard Time",
    "Asia/Kamchatka": "Russia Time Zone 11",
    "Asia/Karachi": "Pakistan Standard Time",
    "Asia/Kashgar": "Central Asia Standard Time",
    "Asia/Katmandu": "Nepal Standard Time",
    "Asia/Khandyga": "Yakutsk Standard Time",
    "Asia/Krasnoyarsk": "North Asia Standard Time",
    "Asia/Kuala_Lumpur": "Singapore Standard Time",
    "Asia/Kuching": "Singapore Standard Time",
    "Asia/Kuwait": "Arab Standard Time",
    "Asia/Macao": "China Standard Time",
    "Asia/Macau": "China Standard Time",
    "Asia/Magadan": "Magadan Standard Time",
    "Asia/Makassar": "Singapore Standard Time",
    "Asia/Manila": "Singapore Standard Time",
    "Asia/Muscat": "Arabian Standard Time",
    "Asia/Nicosia": "GTB Standard Time",
    "Asia/Novokuznetsk": "North Asia Standard Time",
    "Asia/Novosibirsk": "N. Central Asia Standard Time",
    "Asia/Omsk": "Omsk Standard Time",
    "Asia/Oral": "West Asia Standard Time",
    "Asia/Phnom_Penh": "SE Asia Standard Time",
    "Asia/Pontianak": "SE Asia Standard Time",
    "Asia/Pyongyang": "North Korea Standard Time",
    "Asia/Qatar": "Arab Standard Time",
    "Asia/Qostanay": "Central Asia Standard Time",
    "Asia/Qyzylorda": "Qyzylorda Standard Time",
    "Asia/Rangoon": "Myanmar Standard Time",
    "Asia/Riyadh": "Arab Standard Time",
    "Asia/Saigon": "SE Asia Standard Time",
    "Asia/Sakhalin": "Sakhalin Standard Time",
    "Asia/Samarkand": "West Asia Standard Time",
    "Asia/Seoul": "Korea Standard Time",
    "Asia/Shanghai": "China Standard Time",
    "Asia/Singapore": "Singapore Standard Time",
    "Asia/Srednekolymsk": "Russia Time Zone 10",
    "Asia/Taipei": "Taipei Standard Time",
    "Asia/Tashkent": "West Asia Standard Time",
    "Asia/Tbilisi": "Georgian Standard Time",
    "Asia/Tehran": "Iran Standard Time",
    "Asia/Tel_Aviv": "Israel Standard Time",
    "Asia/Thimbu": "Bangladesh Standard Time",
    "Asia/Thimphu": "Bangladesh Standard Time",
    "Asia/Tokyo": "Tokyo Standard Time",
    "Asia/Tomsk": "Tomsk Standard Time",
    "Asia/Ujung_Pandang": "Singapore Standard Time",
    "Asia/Ulaanbaatar": "Ulaanbaatar Standard Time",
    "Asia/Ulan_Bator": "Ulaanbaatar Standard Time",
    "Asia/Urumqi": "Central Asia Standard Time",
    "Asia/Ust-Nera": "Vladivostok Standard Time",
    "Asia/Vientiane": "SE Asia Standard Time",
    "Asia/Vladivostok": "Vladivostok Standard Time",
    "Asia/Yakutsk": "Yakutsk Standard Time",
    "Asia/Yekaterinburg": "Ekaterinburg Standard Time",
    "Asia/Yerevan": "Caucasus Standard Time",
    "Atlantic/Azores": "Azores Standard Time",
    "Atlantic/Bermuda": "Atlantic Standard Time",
    "Atlantic/Canary": "GMT Standard Time",
    "Atlantic/Cape_Verde": "Cape Verde Standard Time",
    "Atlantic/Faeroe": "GMT Standard Time",
    "Atlantic/Jan_Mayen": "W. Europe Standard Time",
    "Atlantic/Madeira": "GMT Standard Time",
    "Atlantic/Reykjavik": "Greenwich Standard Time",
    "Atlantic/South_Georgia": "UTC-02",
    "Atlantic/St_Helena": "Greenwich Standard Time",
    "Atlantic/Stanley": "SA Eastern Standard Time",
    "Australia/ACT": "AUS Eastern Standard Time",
    "Australia/Adelaide": "Cen. Australia Standard Time",
    "Australia/Brisbane": "E. Australia Standard Time",
    "Australia/Broken_Hill": "Cen. Australia Standard Time",
    "Australia/Canberra": "AUS Eastern Standard Time",
    "Australia/Currie": "Tasmania Standard Time",
    "Australia/Darwin": "AUS Central Standard Time",
    "Australia/Eucla": "Aus Central W. Standard Time",
    "Australia/Hobart": "Tasmania Standard Time",
    "Australia/LHI": "Lord Howe Standard Time",
    "Australia/Lindeman": "E. Australia Standard Time",
    "Australia/Lord_Howe": "Lord Howe Standard Time",
    "Australia/Melbourne": "AUS Eastern Standard Time",
    "Australia/NSW": "AUS Eastern Standard Time",
    "Australia/North": "AUS Central Standard Time",
    "Australia/Perth": "W. Australia Standard Time",
    "Australia/Queensland": "E. Australia Standard Time",
    "Australia/South": "Cen. Australia Standard Time",
    "Australia/Sydney": "AUS Eastern Standard Time",
    "Australia/Tasmania": "Tasmania Standard Time",
    "Australia/Victoria": "AUS Eastern Standard Time",
    "Australia/West": "W. Australia Standard Time",
    "Australia/Yancowinna": "Cen. Australia Standard Time",
    "Brazil/Acre": "SA Pacific Standard Time",
    "Brazil/DeNoronha": "UTC-02",
    "Brazil/East": "E. South America Standard Time",
    "Brazil/West": "SA Western Standard Time",
    "CST6CDT": "Central Standard Time",
    "Canada/Atlantic": "Atlantic Standard Time",
    "Canada/Central": "Central Standard Time",
    "Canada/Eastern": "Eastern Standard Time",
    "Canada/Mountain": "Mountain Standard Time",
    "Canada/Newfoundland": "Newfoundland Standard Time",
    "Canada/Pacific": "Pacific Standard Time",
    "Canada/Saskatchewan": "Canada Central Standard Time",
    "Canada/Yukon": "Yukon Standard Time",
    "Chile/Continental": "Pacific SA Standard Time",
    "Chile/EasterIsland": "Easter Island Standard Time",
    "Cuba": "Cuba Standard Time",
    "EST5EDT": "Eastern Standard Time",
    "Egypt": "Egypt Standard Time",
    "Eire": "GMT Standard Time",
    "Etc/GMT": "UTC",
    "Etc/GMT+0": "UTC",
    "Etc/GMT+1": "Cape Verde Standard Time",
    "Etc/GMT+10": "Hawaiian Standard Time",
    "Etc/GMT+11": "UTC-11",
    "Etc/GMT+12": "Dateline Standard Time",
    "Etc/GMT+2": "UTC-02",
    "Etc/GMT+3": "SA Eastern Standard Time",
    "Etc/GMT+4": "SA Western Standard Time",
    "Etc/GMT+5": "SA Pacific Standard Time",
    "Etc/GMT+6": "Central America Standard Time",
    "Etc/GMT+7": "US Mountain Standard Time",
    "Etc/GMT+8": "UTC-08",
    "Etc/GMT+9": "UTC-09",
    "Etc/GMT-0": "UTC",
    "Etc/GMT-1": "W. Central Africa Standard Time",
    "Etc/GMT-10": "West Pacific Standard Time",
    "Etc/GMT-11": "Central Pacific Standard Time",
    "Etc/GMT-12": "UTC+12",
    "Etc/GMT-13": "UTC+13",
    "Etc/GMT-14": "Line Islands Standard Time",
    "Etc/GMT-2": "South Africa Standard Time",
    "Etc/GMT-3": "E. Africa Standard Time",
    "Etc/GMT-4": "Arabian Standard Time",
    "Etc/GMT-5": "West Asia Standard Time",
    "Etc/GMT-6": "Central Asia Standard Time",
    "Etc/GMT-7": "SE Asia Standard Time",
    "Etc/GMT-8": "Singapore Standard Time",
    "Etc/GMT-9": "Tokyo Standard Time",
    "Etc/GMT0": "UTC",
    "Etc/Greenwich": "UTC",
    "Etc/UCT": "UTC",
    "Etc/UTC": "UTC",
    "Etc/Universal": "UTC",
    "Etc/Zulu": "UTC",
    "Europe/Amsterdam": "Romance Standard Time",
    "Europe/Andorra": "W. Europe Standard Time",
    "Europe/Astrakhan": "Astrakhan Standard Time",
    "Europe/Athens": "GTB Standard Time",
    "Europe/Belfast": "GMT Standard Time",
    "Europe/Belgrade": "Central Europe Standard Time",
    "Europe/Berlin": "W. Europe Standard Time",
    "Europe/Bratislava": "Central Europe Standard Time",
    "Europe/Brussels": "Romance Standard Time",
    "Europe/Bucharest": "GTB Standard Time",
    "Europe/Budapest": "Central Europe Standard Time",
    "Europe/Busingen": "W. Europe Standard Time",
    "Europe/Chisinau": "E. Europe Standard Time",
    "Europe/Copenhagen": "W. Europe Standard Time",
    "Europe/Dublin": "GMT Standard Time",
    "Europe/Gibraltar": "W. Europe Standard Time",
    "Europe/Guernsey": "GMT Standard Time",
    "Europe/Helsinki": "FLE Standard Time",
    "Europe/Isle_of_Man": "GMT Standard Time",
    "Europe/Istanbul": "Turkey Standard Time",
    "Europe/Jersey": "GMT Standard Time",
    "Europe/Kaliningrad": "Kaliningrad Standard Time",
    "Europe/Kiev": "FLE Standard Time",
    "Europe/Kirov": "Russian Standard Time",
    "Europe/Lisbon": "GMT Standard Time",
    "Europe/Ljubljana": "Central Europe Standard Time",
    "Europe/London": "GMT Standard Time",
    "Europe/Luxembourg": "Romance Standard Time",
    "Europe/Madrid": "Romance Standard Time",
    "Europe/Malta": "W. Europe Standard Time",
    "Europe/Mariehamn": "FLE Standard Time",
    "Europe/Minsk": "Belarus Standard Time",
    "Europe/Monaco": "Romance Standard Time",
    "Europe/Moscow": "Russian Standard Time",
    "Europe/Nicosia": "GTB Standard Time",
    "Europe/Oslo": "W. Europe Standard Time",
    "Europe/Paris": "Romance Standard Time",
    "Europe/Podgorica": "Central Europe Standard Time",
    "Europe/Prague": "Central Europe Standard Time",
    "Europe/Riga": "FLE Standard Time",
    "Europe/Rome": "W. Europe Standard Time",
    "Europe/Samara": "Russia Time Zone 3",
    "Europe/San_Marino": "W. Europe Standard Time",
    "Europe/Sarajevo": "Central Europe Standard Time",
    "Europe/Saratov": "Saratov Standard Time",
    "Europe/Simferopol": "Russian Standard Time",
    "Europe/Skopje": "Central Europe Standard Time",
    "Europe/Sofia": "FLE Standard Time",
    "Europe/Stockholm": "W. Europe Standard Time",
    "Europe/Tallinn": "FLE Standard Time",
    "Europe/Tirane": "Central Europe Standard Time",
    "Europe/Tiraspol": "E. Europe Standard Time",
    "Europe/Ulyanovsk": "Astrakhan Standard Time",
    "Europe/Uzhgorod": "FLE Standard Time",
    "Europe/Vaduz": "W. Europe Standard Time",
    "Europe/Vatican": "W. Europe Standard Time",
    "Europe/Vienna": "W. Europe Standard Time",
    "Europe/Vilnius": "FLE Standard Time",
    "Europe/Volgograd": "Volgograd Standard Time",
    "Europe/Warsaw": "Central European Standard Time",
    "Europe/Zagreb": "Central Europe Standard Time",
    "Europe/Zaporozhye": "FLE Standard Time",
    "Europe/Zurich": "W. Europe Standard Time",
    "GB": "GMT Standard Time",
    "GB-Eire": "GMT Standard Time",
    "GMT+0": "UTC",
    "GMT-0": "UTC",
    "GMT0": "UTC",
    "Greenwich": "UTC",
    "Hongkong": "China Standard Time",
    "Iceland": "Greenwich Standard Time",
    "Indian/Antananarivo": "E. Africa Standard Time",
    "Indian/Chagos": "Central Asia Standard Time",
    "Indian/Christmas": "SE Asia Standard Time",
    "Indian/Cocos": "Myanmar Standard Time",
    "Indian/Comoro": "E. Africa Standard Time",
    "Indian/Kerguelen": "West Asia Standard Time",
    "Indian/Mahe": "Arabian Standard Time",
    "Indian/Maldives": "West Asia Standard Time",
    "Indian/Mauritius": "Mauritius Standard Time",
    "Indian/Mayotte": "E. Africa Standard Time",
    "Indian/Reunion": "Arabian Standard Time",
    "Iran": "Iran Standard Time",
    "Israel": "Israel Standard Time",
    "Jamaica": "SA Pacific Standard Time",
    "Japan": "Tokyo Standard Time",
    "Kwajalein": "UTC+12",
    "Libya": "Libya Standard Time",
    "MST7MDT": "Mountain Standard Time",
    "Mexico/BajaNorte": "Pacific Standard Time (Mexico)",
    "Mexico/BajaSur": "Mountain Standard Time (Mexico)",
    "Mexico/General": "Central Standard Time (Mexico)",
    "NZ": "New Zealand Standard Time",
    "NZ-CHAT": "Chatham Islands Standard Time",
    "Navajo": "Mountain Standard Time",
    "PRC": "China Standard Time",
    "PST8PDT": "Pacific Standard Time",
    "Pacific/Apia": "Samoa Standard Time",
    "Pacific/Auckland": "New Zealand Standard Time",
    "Pacific/Bougainville": "Bougainville Standard Time",
    "Pacific/Chatham": "Chatham Islands Standard Time",
    "Pacific/Chuuk": "West Pacific Standard Time",
    "Pacific/Easter": "Easter Island Standard Time",
    "Pacific/Efate": "Central Pacific Standard Time",
    "Pacific/Enderbury": "UTC+13",
    "Pacific/Fakaofo": "UTC+13",
    "Pacific/Fiji": "Fiji Standard Time",
    "Pacific/Funafuti": "UTC+12",
    "Pacific/Galapagos": "Central America Standard Time",
    "Pacific/Gambier": "UTC-09",
    "Pacific/Guadalcanal": "Central Pacific Standard Time",
    "Pacific/Guam": "West Pacific Standard Time",
    "Pacific/Honolulu": "Hawaiian Standard Time",
    "Pacific/Johnston": "Hawaiian Standard Time",
    "Pacific/Kiritimati": "Line Islands Standard Time",
    "Pacific/Kosrae": "Central Pacific Standard Time",
    "Pacific/Kwajalein": "UTC+12",
    "Pacific/Majuro": "UTC+12",
    "Pacific/Marquesas": "Marquesas Standard Time",
    "Pacific/Midway": "UTC-11",
    "Pacific/Nauru": "UTC+12",
    "Pacific/Niue": "UTC-11",
    "Pacific/Norfolk": "Norfolk Standard Time",
    "Pacific/Noumea": "Central Pacific Standard Time",
    "Pacific/Pago_Pago": "UTC-11",
    "Pacific/Palau": "Tokyo Standard Time",
    "Pacific/Pitcairn": "UTC-08",
    "Pacific/Pohnpei": "Central Pacific Standard Time",
    "Pacific/Ponape": "Central Pacific Standard Time",
    "Pacific/Port_Moresby": "West Pacific Standard Time",
    "Pacific/Rarotonga": "Hawaiian Standard Time",
    "Pacific/Saipan": "West Pacific Standard Time",
    "Pacific/Samoa": "UTC-11",
    "Pacific/Tahiti": "Hawaiian Standard Time",
    "Pacific/Tarawa": "UTC+12",
    "Pacific/Tongatapu": "Tonga Standard Time",
    "Pacific/Truk": "West Pacific Standard Time",
    "Pacific/Wake": "UTC+12",
    "Pacific/Wallis": "UTC+12",
    "Pacific/Yap": "West Pacific Standard Time",
    "Poland": "Central European Standard Time",
    "Portugal": "GMT Standard Time",
    "ROC": "Taipei Standard Time",
    "ROK": "Korea Standard Time",
    "Singapore": "Singapore Standard Time",
    "Turkey": "Turkey Standard Time",
    "UCT": "UTC",
    "US/Alaska": "Alaskan Standard Time",
    "US/Aleutian": "Aleutian Standard Time",
    "US/Arizona": "US Mountain Standard Time",
    "US/Central": "Central Standard Time",
    "US/Eastern": "Eastern Standard Time",
    "US/Hawaii": "Hawaiian Standard Time",
    "US/Indiana-Starke": "Central Standard Time",
    "US/Michigan": "Eastern Standard Time",
    "US/Mountain": "Mountain Standard Time",
    "US/Pacific": "Pacific Standard Time",
    "US/Samoa": "UTC-11",
    "UTC": "UTC",
    "Universal": "UTC",
    "W-SU": "Russian Standard Time",
    "Zulu": "UTC",
}



# PAHO
# TODO: implement config validation and base config format
# TODO: add queue for to_send messages

# TODO: adding a lock for accessing self._mqttc should solve some of the bugs, but it introduces a new one
# basically, when a user thread calls send, they should acquire the lock for the self._mqttc object
# and use it to send messages. However, if the mqttc has loop started but did not connect, the lock will
# prevent the client from ever connecting.

import os
import traceback
from collections import deque
from threading import Lock
from time import sleep

from ..const import BASE_CT, COLORS, COMMS, PAYLOAD_CT
from ..utils import resolve_domain_or_ip

from importlib import resources as impresources
from .. import certs


class BaseCommWrapper(object):
  def __init__(
      self,
      log,
      config,
      recv_buff=None,
      send_channel_name=None,
      recv_channel_name=None,
      comm_type=None,
      verbosity=1,
      **kwargs
  ):
    self.log = log
    self._config = config
    self._recv_buff = recv_buff
    self._send_to = None
    self.send_channel_name = send_channel_name
    self.recv_channel_name = recv_channel_name
    self._comm_type = comm_type
    self.__verbosity = verbosity
    super(BaseCommWrapper, self).__init__(**kwargs)
    return

  @property
  def comm_log_prefix(self):
    return 'COMMWRP'

  @property
  def verbosity(self):
    return self.__verbosity

  def P(self, s, color=None, verbosity=1, **kwargs):
    if verbosity > self.verbosity:
      return
    if color is None or (isinstance(color, str) and color[0] not in ['e', 'r']):
      color = COLORS.COMM
    comtype = self._comm_type[:7] if self._comm_type is not None else 'CUSTOM'
    self.log.P(f"[{self.comm_log_prefix}][{comtype}] {s}", color=color, **kwargs)
    return

  def D(self, s, t=False):
    _r = -1
    if self.DEBUG:
      if self.show_prefixes:
        msg = "[DEBUG] {}: {}".format(self.__name__, s)
      else:
        if self.prefix_log is None:
          msg = "[D] {}".format(s)
        else:
          msg = "[D]{} {}".format(self.prefix_log, s)
        # endif
      # endif
      _r = self.log.P(msg, show_time=t, color='yellow')
    # endif
    return _r

  @property
  def is_secured(self):
    val = self.cfg_secured
    if isinstance(val, str):
      val = val.upper() in ["1", "TRUE", "YES"]
    return val

  @property
  def send_channel_name(self):
    return self._send_channel_name

  @property
  def recv_channel_name(self):
    return self._recv_channel_name

  @send_channel_name.setter
  def send_channel_name(self, x):
    if isinstance(x, tuple):
      self._send_channel_name, self._send_to = x
    else:
      self._send_channel_name = x
    return

  @recv_channel_name.setter
  def recv_channel_name(self, x):
    self._recv_channel_name = x
    return

  @property
  def cfg_node_id(self):
    return self._config.get(COMMS.EE_ID, self._config.get(COMMS.SB_ID, None))

  @property
  def cfg_node_addr(self):
    return self._config.get(COMMS.EE_ADDR)

  @property
  def cfg_user(self):
    return self._config[COMMS.USER]

  @property
  def cfg_pass(self):
    return self._config[COMMS.PASS]

  @property
  def cfg_host(self):
    return self._config[COMMS.HOST]

  @property
  def cfg_port(self):
    return self._config[COMMS.PORT]

  @property
  def cfg_qos(self):
    return self._config[COMMS.QOS]

  @property
  def cfg_cert_path(self):
    return self._config.get(COMMS.CERT_PATH)

  @property
  def cfg_secured(self):
    return self._config.get(COMMS.SECURED, 0)  # TODO: make 1 later on

  @property
  def cfg_subtopic(self):
    return self._config.get(COMMS.SUBTOPIC, COMMS.DEFAULT_SUBTOPIC_VALUE)

  def get_subtopic_values(self):
    if self.cfg_subtopic == 'alias':
      # This is done in order for alias to still work with addresses
      return [self.cfg_node_id, self.cfg_node_addr]
    return [self.cfg_node_addr]

  @property
  def channel_key(self):
    """
    Returns the key for the channel in the configuration
    Returns
    -------
    str : The key for the channel in the configuration
    """
    return COMMS.TOPIC

  def extract_channel_from_config(self, cfg):
    """
    Utility function to extract the channel from the configuration
    Parameters
    ----------
    cfg : dict
      The configuration dictionary

    Returns
    -------
    str : The channel
    """
    return cfg[COMMS.TOPIC]

  def get_send_channel_def(self, send_to=None):
    """
    Returns the channel definition for the sender
    Parameters
    ----------
    send_to : str, optional

    Returns
    -------
    dict or None : The channel definition or None if the channel is not defined.
    """
    if self.send_channel_name is None:
      return

    cfg = self._config[self.send_channel_name].copy()
    extracted_channel = self.extract_channel_from_config(cfg)
    if "{}" in extracted_channel:
      if send_to is not None:
        extracted_channel = extracted_channel.format(send_to)
      # TODO: ._send_to seems to be necessary only for the amqp wrapper
      elif self._send_to is not None:
        extracted_channel = extracted_channel.format(self._send_to)
    # endif check for {}

    assert "{}" not in extracted_channel

    cfg[self.channel_key] = extracted_channel
    return cfg

  def server_connect(self, max_retries=5):
    raise NotImplementedError

  def send(self, message):
    raise NotImplementedError

  def receive(self):
    raise NotImplementedError

  def release(self):
    raise NotImplementedError



class EvmNetData:
  DAUTH_URL_KEY = 'EE_DAUTH_URL'
  DAUTH_ND_ADDR_KEY = 'EE_DAUTH_ND_ADDR'
  DAUTH_RPC_KEY = 'EE_DAUTH_RPC'
  DAUTH_R1_ADDR_KEY = 'EE_DAUTH_R1_ADDR'
  DAUTH_MND_ADDR_KEY = 'EE_DAUTH_MND_ADDR'
  DAUTH_PROXYAPI_ADDR_KEY = 'EE_DAUTH_PROXYAPI_ADDR'
  
  DAUTH_CONTROLLER_ADDR_KEY = 'EE_DAUTH_CONTROLLER_ADDR'  
  DAUTH_GET_ORACLES_ABI = 'EE_DAUTH_GET_ORACLES_ABI'
  
  EE_GENESIS_EPOCH_DATE_KEY = 'EE_GENESIS_EPOCH_DATE'
  EE_EPOCH_INTERVALS_KEY = 'EE_EPOCH_INTERVALS'
  EE_EPOCH_INTERVAL_SECONDS_KEY = 'EE_EPOCH_INTERVAL_SECONDS'
  
  EE_SUPERVISOR_MIN_AVAIL_PRC_KEY = 'EE_SUPERVISOR_MIN_AVAIL_PRC'

  EE_ORACLE_API_URL_KEY = 'EE_ORACLE_API_URL'
  

_DAUTH_ABI_IS_NODE_ACTIVE = [{
  "inputs": [
    {
      "internalType": "address",
      "name": "nodeAddress",
      "type": "address"
    }
  ],
  "name": "isNodeActive",
  "outputs": [
    {
      "internalType": "bool",
      "name": "",
      "type": "bool"
    }
  ],
  "stateMutability": "view",
  "type": "function"
}]


_DAUTH_ABI_GET_ORACLES = [{
  "inputs": [],
  "name": "getOracles",
  "outputs": [
    {
      "internalType": "address[]",
      "name": "",
      "type": "address[]"
    }
  ],
  "stateMutability": "view",
  "type": "function"
}]


_DAUTH_ABI_GET_SIGNERS_deprecated = [{
  "inputs": [],
  "name": "getSigners",
  "outputs": [
    {
      "internalType": "address[]",
      "name": "",
      "type": "address[]"
    }
  ],
  "stateMutability": "view",
  "type": "function"
}]


_GET_NODE_INFO_ABI = [
  {
      "inputs": [
        {
          "internalType": "address",
          "name": "node",
          "type": "address"
        }
      ],
      "name": "getNodeLicenseDetails",
      "outputs": [
        {
          "components": [
            {
              "internalType": "enum LicenseType",
              "name": "licenseType",
              "type": "uint8"
            },
            {
              "internalType": "uint256",
              "name": "licenseId",
              "type": "uint256"
            },
            {
              "internalType": "address",
              "name": "owner",
              "type": "address"
            },
            {
              "internalType": "address",
              "name": "nodeAddress",
              "type": "address"
            },
            {
              "internalType": "uint256",
              "name": "totalAssignedAmount",
              "type": "uint256"
            },
            {
              "internalType": "uint256",
              "name": "totalClaimedAmount",
              "type": "uint256"
            },
            {
              "internalType": "uint256",
              "name": "lastClaimEpoch",
              "type": "uint256"
            },
            {
              "internalType": "uint256",
              "name": "assignTimestamp",
              "type": "uint256"
            },
            {
              "internalType": "address",
              "name": "lastClaimOracle",
              "type": "address"
            },
            {
              "internalType": "bool",
              "name": "isBanned",
              "type": "bool"
            }
          ],
          "internalType": "struct LicenseDetails",
          "name": "",
          "type": "tuple"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    }
]


_GET_WALLET_NODES = [
    {
      "inputs": [
        {
          "internalType": "address",
          "name": "wallet",
          "type": "address"
        }
      ],
      "name": "getWalletNodes",
      "outputs": [
        {
          "internalType": "address[]",
          "name": "nodes",
          "type": "address[]"
        }
      ],
      "stateMutability": "view",
      "type": "function"
    },
                  
]

# A minimal ERC20 ABI for balanceOf, transfer, and decimals functions.
_ERC20_ABI = [
  {
      "constant": True,
      "inputs": [{"name": "_owner", "type": "address"}],
      "name": "balanceOf",
      "outputs": [{"name": "balance", "type": "uint256"}],
      "payable": False,
      "stateMutability": "view",
      "type": "function"
  },
  {
      "constant": False,
      "inputs": [
          {"name": "_to", "type": "address"},
          {"name": "_value", "type": "uint256"}
      ],
      "name": "transfer",
      "outputs": [{"name": "success", "type": "bool"}],
      "payable": False,
      "stateMutability": "nonpayable",
      "type": "function"
  },
  {
      "constant": True,
      "inputs": [],
      "name": "decimals",
      "outputs": [{"name": "", "type": "uint8"}],
      "payable": False,
      "stateMutability": "view",
      "type": "function"
  }
]


  
  
EVM_NET_DATA = {
  'mainnet': {
    EvmNetData.DAUTH_URL_KEY                    : "https://dauth.ratio1.ai/get_auth_data",
    EvmNetData.DAUTH_CONTROLLER_ADDR_KEY        : "0x90dA5FdaA92edDC80FB73114fb7FE7D97f2be017",
    EvmNetData.DAUTH_ND_ADDR_KEY                : "0xE658DF6dA3FB5d4FBa562F1D5934bd0F9c6bd423",
    EvmNetData.DAUTH_R1_ADDR_KEY                : "0x6444C6c2D527D85EA97032da9A7504d6d1448ecF",
    EvmNetData.DAUTH_MND_ADDR_KEY               : "0x0C431e546371C87354714Fcc1a13365391A549E2",
    EvmNetData.DAUTH_PROXYAPI_ADDR_KEY          : "0x0fC093d5f4B7a3fb752884397F4878f097E5Be1E",
    EvmNetData.DAUTH_RPC_KEY                    : "https://base-mainnet.public.blastapi.io",
    EvmNetData.EE_GENESIS_EPOCH_DATE_KEY        : "2025-05-23 16:00:00",
    EvmNetData.EE_EPOCH_INTERVALS_KEY           : 24,
    EvmNetData.EE_EPOCH_INTERVAL_SECONDS_KEY    : 3600,
    EvmNetData.EE_SUPERVISOR_MIN_AVAIL_PRC_KEY  : 0.98,
    EvmNetData.EE_ORACLE_API_URL_KEY            : "https://oracle.ratio1.ai",
    EvmNetData.DAUTH_GET_ORACLES_ABI            : _DAUTH_ABI_GET_ORACLES,
  },

  'testnet': {
    EvmNetData.DAUTH_URL_KEY                    : "https://testnet-dauth.ratio1.ai/get_auth_data",
    EvmNetData.DAUTH_CONTROLLER_ADDR_KEY        : "0x63BEC1B3004154698830C7736107E7d3cfcbde79",
    EvmNetData.DAUTH_ND_ADDR_KEY                : "0x18E86a5829CA1F02226FA123f30d90dCd7cFd0ED",
    EvmNetData.DAUTH_R1_ADDR_KEY                : "0xCC96f389F45Fc08b4fa8e2bC4C7DA9920292ec64",
    EvmNetData.DAUTH_MND_ADDR_KEY               : "0xa8d7FFCE91a888872A9f5431B4Dd6c0c135055c1",
    EvmNetData.DAUTH_PROXYAPI_ADDR_KEY          : "0xF94b53855Fde16cbF3f9C3e300e2E6A495AE0A0A",
    EvmNetData.DAUTH_RPC_KEY                    : "https://base-sepolia.public.blastapi.io",      
    EvmNetData.EE_GENESIS_EPOCH_DATE_KEY        : "2025-05-23 16:00:00",
    EvmNetData.EE_EPOCH_INTERVALS_KEY           : 24,
    EvmNetData.EE_EPOCH_INTERVAL_SECONDS_KEY    : 3600,
    EvmNetData.EE_SUPERVISOR_MIN_AVAIL_PRC_KEY  : 0.6,
    EvmNetData.EE_ORACLE_API_URL_KEY            : "https://testnet-oracle.ratio1.ai",
    EvmNetData.DAUTH_GET_ORACLES_ABI            : _DAUTH_ABI_GET_ORACLES,
  },

  
  'devnet' : {
    EvmNetData.DAUTH_URL_KEY                    : "https://devnet-dauth.ratio1.ai/get_auth_data",
    EvmNetData.DAUTH_CONTROLLER_ADDR_KEY        : "0xdd56E920810e2FD9a07C1718643E179839867253",
    EvmNetData.DAUTH_ND_ADDR_KEY                : "0x8D0CE4933728FF7C04388f0bEcC9a45676E232F7",
    EvmNetData.DAUTH_R1_ADDR_KEY                : "0x07C5678F0f4aC347496eAA8D6031b37FF3402CE5",
    EvmNetData.DAUTH_MND_ADDR_KEY               : "0x7A14Be75135a7ebdef99339CCc700C25Cda60c6E",
    EvmNetData.DAUTH_PROXYAPI_ADDR_KEY          : "0x746aaB0a8bcFB92094Acc371D7D6A2F69DaA23E3",
    EvmNetData.DAUTH_RPC_KEY                    : "https://base-sepolia.public.blastapi.io",
    EvmNetData.EE_GENESIS_EPOCH_DATE_KEY        : "2025-05-23 16:00:00",
    EvmNetData.EE_EPOCH_INTERVALS_KEY           : 1,
    EvmNetData.EE_EPOCH_INTERVAL_SECONDS_KEY    : 3600,
    EvmNetData.EE_SUPERVISOR_MIN_AVAIL_PRC_KEY  : 0.6,
    EvmNetData.EE_ORACLE_API_URL_KEY            : "https://devnet-oracle.ratio1.ai",
    EvmNetData.DAUTH_GET_ORACLES_ABI            : _DAUTH_ABI_GET_ORACLES,
  },
}

class EVM_ABI_DATA:
  GET_NODE_INFO = _GET_NODE_INFO_ABI
  GET_WALLET_NODES = _GET_WALLET_NODES
  IS_NODE_ACTIVE = _DAUTH_ABI_IS_NODE_ACTIVE
  ERC20_ABI = _ERC20_ABI
  


# local dependencies
from ...io_formatter.base import BaseFormatter


class Aixp1Formatter(BaseFormatter):

  def __init__(self, log, **kwargs):
    super(Aixp1Formatter, self).__init__(
        log=log, prefix_log='[INV-FORM]', **kwargs)
    return

  def startup(self):
    pass

  def _encode_output(self, output):
    event_type = output.pop('EE_EVENT_TYPE', None)

    # below fields are not required as they will be decorated post-formatting anyway
    output.pop('EE_MESSAGE_ID', None)
    output.pop('EE_MESSAGE_SEQ', None)
    output.pop('EE_TOTAL_MESSAGES', None)

    output.pop('EE_TIMESTAMP', None)
    output.pop('EE_ID', None)
    output.pop('STREAM_NAME', None)
    output.pop('SIGNATURE', None)
    output.pop('INSTANCE_ID', None)

    output.pop('EE_TIMEZONE', None)
    output.pop('EE_VERSION', None)
    output.pop('EE_TZ', None)

    output.pop('INITIATOR_ID', None)
    output.pop('SESSION_ID', None)
    # end non-managed fields

    lvl_0_dct = {
      "DATA": {
      },
    }

    lvl_1_dct = lvl_0_dct['DATA']

    if event_type == 'PAYLOAD':
      # add payload context
      output.pop('STREAM')
      output.pop('PIPELINE')

      # Plugin meta
      if True:
        lvl_1_dct['PLUGIN_META'] = {}
        plugin_meta_keys = [k for k in output.keys() if k.startswith('_P_')]
        for k in plugin_meta_keys:
          lvl_1_dct['PLUGIN_META'][k] = output.pop(k, None)
        # end for

      # Pipeline meta
      if True:
        lvl_1_dct['PIPELINE_META'] = {}
        pipeline_meta_keys = [k for k in output.keys() if k.startswith('_C_')]
        for k in pipeline_meta_keys:
          lvl_1_dct['PIPELINE_META'][k] = output.pop(k, None)
        # end for

    # endif payload

    for k, v in output.items():
      lvl_1_dct[k] = v

    return lvl_0_dct

  def _decode_output(self, encoded_output):
    # Pop the unimportant stuff
    encoded_output.get('EE_FORMATTER', None)

    node_id, pipeline, signature, instance_id = encoded_output.get('EE_PAYLOAD_PATH', [None, None, None, None])

    encoded_output['EE_ID'] = node_id

    if encoded_output['EE_EVENT_TYPE'] != 'HEARTBEAT':
      encoded_output['STREAM_NAME'] = pipeline

    if pipeline is not None:
      encoded_output['SIGNATURE'] = signature

    if instance_id is not None:
      encoded_output['INSTANCE_ID'] = instance_id

    lvl_1_dct = encoded_output.pop('DATA')
    if encoded_output['EE_EVENT_TYPE'] == 'PAYLOAD':
      encoded_output['STREAM'] = pipeline
      encoded_output['PIPELINE'] = pipeline

      # Plugin meta
      plugin_meta = lvl_1_dct.pop('PLUGIN_META', {}) or {}
      for k, v in plugin_meta.items():
        encoded_output[k] = v

      # Pipeline meta
      pipeline_meta = lvl_1_dct.pop('PIPELINE_META', {}) or {}
      for k, v in pipeline_meta.items():
        encoded_output[k] = v

      encoded_output['PLUGIN_CATEGORY'] = 'general'
      a = 1

    for k, v in lvl_1_dct.items():
      encoded_output[k] = v

    return encoded_output

  def _decode_streams(self, dct_config_streams):
    return dct_config_streams



"""
ex18_deploy_container_webapi.py
---------------------

This is a basic example of how to use the ratio1 SDK for a containerized web API deployment

The example further extends ex02_part1_deploy_webapi.py and shows how to deploy a containerized web API rather than
a web API that uses Ratio1 Edge Node shallow machine learning internal (API) capabilities.

"""
from ratio1 import Session, CustomPluginTemplate, PLUGIN_TYPES

if __name__ == '__main__':
  
  session = Session(silent=True)

  my_node = "0xai_ApM1AbzLq1VtsLIidmvzt1Nv4Cyl5Wed0fHNMoZv9u4X"

  # NOTE: When working with SDK please use the nodes internal addresses. While the EVM address of the node
  #       is basically based on the same sk/pk it is in a different format and not directly usable with the SDK
  #       the internal node address is easily spoted as starting with 0xai_ and can be found 
  #       via `docker exec r1node get_node_info` or via the launcher UI
  
  app, _ = session.create_container_web_app( # this uses CONTAINER_APP_RUNNER
    node=my_node,
    name="ratio1_simple_container_webapp",
    image="tvitalii/flask-docker-app:latest",
    port=5000,
    # container_resources={},
    # cr="docker.io",
  )

  try:
    url = app.deploy()
    print("Webapp deployed at: ", url)
  except Exception as e:
    print("Error deploying webapp: ", e)

  # Observation:
  #   next code is not mandatory - it is used to keep the session open and cleanup the resources
  #   due to the fact that this is a example/tutorial and maybe we dont want to keep the pipeline
  #   active after the session is closed we use close_pipelines=True
  #   in production, you would not need this code as the script can close 
  #   after the pipeline will be sent 
  session.wait(
    seconds=120,            # we wait the session for 60 seconds
    close_pipelines=True,   # we close the pipelines after the session
    close_session=True,     # we close the session after the session
  )



from ...io_formatter.base import BaseFormatter
import re


def camel_to_upper_snake(name):
  name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
  return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).upper()


def snake_to_camel(name):
  name = name.lower()

  words = []
  for i, x in enumerate(name.split('_')):
    if i > 0:
      words.append(x.title())
    else:
      words.append(x)

  return ''.join(words)


class ADummyFormatter(BaseFormatter):

  def __init__(self, **kwargs):
    super(ADummyFormatter, self).__init__(**kwargs)
    return

  def _decode_streams(self, dct_config_streams):
    return dct_config_streams

  def _encode_output(self, output):
    # transforms the first level keys to camelCase
    new_output = {}

    for k, v in output.items():
      k_camel_case = snake_to_camel(k)
      new_output[k_camel_case] = v

    return new_output

  def _decode_output(self, encoded_output):
    # it is used mainly for map-reduce jobs in reduce phase!
    decoded_output = {}

    for k, v in encoded_output.items():
      k_snake_case = camel_to_upper_snake(k)
      decoded_output[k_snake_case] = v

    return decoded_output



"""
ex02_part1_deploy_webapi.py
---------------------

This is a basic example of how to use the ratio1 SDK for a simple web API deployment that uses Ratio1
Edge Node shallow machine learning internal (API) capabilities.

"""
from ratio1 import Session, CustomPluginTemplate, PLUGIN_TYPES

def run_predict(plugin: CustomPluginTemplate, inputs: list[int], nr_steps: int) -> list:
  """
  Here we use the Ratio1 build in basic ML internal API
  """
  preds = plugin.basic_ts_fit_predict(inputs, nr_steps)
  # now we apply some simple heuristic to get the lower and upper bounds
  lower_bound = []
  upper_bound = []
  heuristic_factor = 0.1
  for pred in preds:
    lower_bound.append(int(pred - abs(pred * heuristic_factor)))
    upper_bound.append(int(pred + abs(pred * heuristic_factor)))
  #endfor
  result = {
    "predictions": [round(x,1) for x in preds],
    "lower_bound": lower_bound,
    "upper_bound": upper_bound
  }
  return result


if __name__ == '__main__':
  
  session = Session(silent=True)

  my_node = "0xai_ApM1AbzLq1VtsLIidmvzt1Nv4Cyl5Wed0fHNMoZv9u4X"

  # NOTE: When working with SDK please use the nodes internal addresses. While the EVM address of the node
  #       is basically based on the same sk/pk it is in a different format and not directly usable with the SDK
  #       the internal node address is easily spoted as starting with 0xai_ and can be found 
  #       via `docker exec r1node get_node_info` or via the launcher UI
  
  app, _ = session.create_custom_webapi(
    node=my_node,
    name="ratio1_simple_predict_webapp",
    endpoints=[
      {
        "function": run_predict,
        "method": "post",
      },        
    ]
  )
  try:
    url = app.deploy()
    print("Webapp deployed at: ", url)
  except Exception as e:
    print("Error deploying webapp: ", e)
  
  # Observation:
  #   next code is not mandatory - it is used to keep the session open and cleanup the resources
  #   due to the fact that this is a example/tutorial and maybe we dont want to keep the pipeline
  #   active after the session is closed we use close_pipelines=True
  #   in production, you would not need this code as the script can close 
  #   after the pipeline will be sent 
  session.wait(
    seconds=120,            # we wait the session for 60 seconds
    close_pipelines=True,   # we close the pipelines after the session
    close_session=True,     # we close the session after the session
  )


import json
import os
from pathlib import Path
import shutil

from ratio1.const.base import BCct, dAuth, EE_SDK_ALIAS_DEFAULT, EE_SDK_ALIAS_ENV_KEY 
from ratio1._ver import __VER__ as version

from ratio1.logging.base_logger import SDK_HOME, BaseLogger


CONFIG_FILE = "config"

LOCAL_PEM_PATH = "./_local_cache/_data/" + BCct.DEFAULT_PEM_FILE

ENV_TEMPLATE = """
# Configuration file for the Ratio1 SDK

EE_EVM_NET=testnet
EE_TARGET_NODE=

"""

def _create_bc_engine():
  from ratio1.bc import DefaultBlockEngine
  from ratio1 import Logger
  return DefaultBlockEngine(
    name="default", 
    log=Logger("CLI", silent=True),
    user_config=True, # this is must to use the user config
  )
  

def seconds_to_short_format(seconds):
  """
  Converts a duration in seconds into a short human-readable format: "Xd HH:MM:SS".

  Parameters
  ----------
  seconds : int
      The total duration in seconds.

  Returns
  -------
  str
      Short human-readable duration in "Xd HH:MM:SS" format.
  """
  days = int(seconds / (24 * 3600))
  seconds %= (24 * 3600)
  hours = int(seconds / 3600)
  seconds %= 3600
  minutes = int(seconds / 60)
  seconds %= 60
  seconds = int(seconds)

  # Format the result
  if days > 0:
    return f"{days}d {hours:02}:{minutes:02}:{seconds:02}"
  else:
    return f"{hours:02}:{minutes:02}:{seconds:02}"

def log_with_color(s, color='n'):
  """
  Prints the string `s` to the console in the specified color.

  Parameters
  ----------
  s : str
      The string to be logged.
  color : str, optional
      The color code: 
      'r' for red, 'g' for green, 'y' for yellow, 
      'b' for blue, 'w' for light white, 'n' for dark white (default).
  """
  color_codes = {
      'r': '\033[31m',  # Red
      'g': '\033[32m',  # Green
      'y': '\033[33m',  # Yellow
      # 'b': "\x1b[1;34m",  # bright blue
      'b': "\033[36m",  # bright cyan
      'w': '\033[97m',  # Light white
      'c': "\033[36m",  # bright cyan
      'n': '\033[37m',  # white (default)
      'd': "\033[90m",  # dark gray
      
  }

  reset_code = '\033[0m'  # Reset color
  color_code = color_codes.get(color, color_codes['n'])
  print(f"{color_code}{s}{reset_code}", flush=True)


def get_user_folder():
  """
  Returns the user folder.
  """
  return BaseLogger.get_user_folder(as_str=False, include_sdk_home=True)


def get_user_config_file():
  """
  Returns the user configuration file.
  """
  return get_user_folder() / CONFIG_FILE


def get_network():
  return os.environ.get(dAuth.DAUTH_NET_ENV_KEY, dAuth.DAUTH_SDK_NET_DEFAULT)

def get_alias():
  return os.environ.get(EE_SDK_ALIAS_ENV_KEY, EE_SDK_ALIAS_DEFAULT)

def set_client_alias(alias : str):
  config_file = get_user_config_file()
  # open config_file and update EE_SDK_ALIAS
  with config_file.open("r") as file:
    lines = file.readlines()
  with config_file.open("w") as file:
    found = False
    for line in lines:
      if line.startswith(EE_SDK_ALIAS_ENV_KEY):
        line = f"{EE_SDK_ALIAS_ENV_KEY}={alias}\n"
        found = True
      file.write(line)
    if not found:
      file.write(f"{EE_SDK_ALIAS_ENV_KEY}={alias}\n")
  log_with_color(f"Alias set to {alias}", color='b')
  return 

def get_networks(args):
  """
  Shows the network configuration.
  """
  log_with_color(f"Client v{version} on network: {get_network()}", color='b')
  from ratio1.const.evm_net import EVM_NET_DATA
  log_with_color(f"Available networks:\n{json.dumps(EVM_NET_DATA, indent=2)}", color='w')  
  return


def get_set_network(args):
  net = args.new or args.set
  env_network = get_network()
  if net is None:
    log_with_color(f"Client v{version} on network: {env_network}", color='b')
  else:
    config_file = get_user_config_file()
    # open config_file and update EE_EVM_NET
    with config_file.open("r") as file:
      lines = file.readlines()
    with config_file.open("w") as file:
      found = False
      for line in lines:
        if line.startswith("EE_EVM_NET"):
          line = f"EE_EVM_NET={net}\n"
          found = True
        file.write(line)
      if not found:
        file.write(f"EE_EVM_NET={net}\n")
    log_with_color(f"Network set to {net}", color='b')
  return

def get_set_alias(args):
  alias = args.set
  if alias is None:
    log_with_color(f"Client v{version} alias: {get_alias()}", color='b')
  else:
    set_client_alias(alias)
  return
    
  

def reset_config(*larg, keep_existing=False, **kwargs):
  """
  Resets the configuration by creating a ~/.ratio1 folder and populating
  ~/.ratio1/config with values from a local .env file, if it exists.
  """
  log_with_color(f"Client v{version} resetting the configuration...", color='y')
  # Define the target config folder and file
  config_dir = get_user_folder()
  config_file = get_user_config_file()
  
  local_pem = Path(LOCAL_PEM_PATH)
  target_pem = config_dir / BCct.USER_PEM_FILE
  
  # Create the ~/.ratio1 folder if it doesn't exist
  config_dir.mkdir(parents=True, exist_ok=True)

  # Check if the current folder has a .env file
  current_env_file = Path(".env")
  if current_env_file.exists():
    # Copy .env content to ~/.ratio1/config
    shutil.copy(current_env_file, config_file)
    log_with_color(
      f"Configuration has been reset using {current_env_file} into {config_file}", 
      color='y'
    )
    log_with_color(f"Please REVIEW the configuration in the file {config_file}", color='b')
  else:
    # Create an empty config file
    with config_file.open("wt") as file:
      file.write(ENV_TEMPLATE)
    log_with_color(
      f"Configuration has been reset to default in {config_file}:\n{ENV_TEMPLATE}", 
      color='y'
    )
    log_with_color(f"Please UPDATE the configuration in the file {config_file}", color='b')
  
  if local_pem.exists():
    log_with_color(f"Copying local PEM file {local_pem} to {target_pem}", color='y')
    shutil.copy(local_pem, target_pem)
  else:
    log_with_color(f"No local PEM file found locally {local_pem}.", color='r')
    if target_pem.exists():
      log_with_color(f"Found already existing {target_pem}.", color='y')
      if not keep_existing:
        target_pem.unlink()        
        log_with_color(f"Deleted {target_pem}. A default key will be generated.", color='b')
      #end if
    #endif pem exists
  #end if local pem exists
  show_version()
  return


def show_address(args):
  """
  Displays the current client address.
  """
  from ratio1 import Session
  sess = Session(
    silent=True
  )
  log_with_color(f"{sess.get_client_address()}", color='b')
  return


def show_version(silent=True):
  from ratio1 import Session
  sess = Session(
    silent=silent
  )  
  
  nodes = sess.get_active_nodes()
  oracles = sess.get_active_supervisors()
  
  user_folder = get_user_folder()  
  
  # TODO: get the epoch from the SDK  - AFTER moving get_epoch_id from core
  
  log_with_color(f"Ratio1 client v{version}:\n", color='b')
  log_with_color(f"SDK folder:     {user_folder}", color='b')
  log_with_color(f"Ratio1 network: {get_network()}", color='b')
  log_with_color(f"Network Epoch:  {sess.bc_engine.get_current_epoch()}", color='b')
  log_with_color(f"SDK addr:       {sess.get_client_address()}", color='b')
  log_with_color(f"SDK ETH addr:   {sess.bc_engine.eth_address}", color='b')
  log_with_color(f"SDK alias:      {sess.name}", color='b')
  log_with_color(f"Active oracles: {len(oracles)}", color='b')
  log_with_color(f"Active nodes:   {len(nodes)}", color='b')
  return


  

def show_config(args):
  """
  Displays the current configuration from ~/.ratio1/config.
  """
  show_version(silent=not args.verbose)
  config_file = get_user_config_file()
  if config_file.exists():
    keys = []
    with config_file.open("r") as file:
      lines = file.readlines()
      keys = [line.strip().split("=")[0] for line in lines if "=" in line and not line.strip().startswith("#")]
      commments = [line.strip() for line in lines if line.strip().startswith("#")]
    log_with_color(f"Current config {config_file} has {len(keys)} keys and {len(commments)} comments\n", color='d')
  else:
    log_with_color(f"No configuration found at {config_file}. Please run `reset_config` first.", color="r")
  return



def load_user_defined_config(verbose=False):
  """
  Loads the ~/.ratio1/config file into the current environment.
  """
  config_file = get_user_config_file()
  result = False
  loaded_keys = []
  if config_file.exists():
    with config_file.open("r") as file:
      for line in file:
        # Ignore comments and empty lines
        if line.strip() and not line.strip().startswith("#"):
          key, value = line.strip().split("=", 1)
          value = value.strip()
          # if at least one key-value pair is found, set the result to True
          if value != "":
            result = True
          os.environ[key.strip()] = value
          loaded_keys.append(key.strip())
    if verbose:
      log_with_color(f"{config_file} loaded into the env: {loaded_keys}", color='b')
  else:
    if verbose:
      log_with_color(f"No configuration file found at {config_file}. Please run `reset_config` first.", color="r")
  return result
  

def maybe_init_config():
  """
  Initializes the configuration if it doesn't exist yet.
  
  TODO: in v3+ this will be migrated to a online authentication process.
  """
  config_file = get_user_config_file()
  if not config_file.exists():
    BaseLogger.maybe_migrate_user_folder()
    if not config_file.exists():
      log_with_color(f"No configuration file found at {config_file}. Initializing configuration...", color="y")
      reset_config(keep_existing=True)
      return False
    # config_file still does not exist even after attempting the migration.
  return load_user_defined_config()



def get_eth_addr(args):
  """
  Gets the ETH address given a node address.
  """
  node = args.node
  eng = _create_bc_engine()
  eth_addr = eng.get_eth_address(node)
  log_with_color(f"ETH address for node {node}: {eth_addr}", color='b')
  return


import os
from copy import deepcopy

from ratio1 import Logger, const
from ratio1.bc import DefaultBlockEngine



if __name__ == '__main__' :
  l = Logger("ENC", base_folder=".", app_folder="_local_cache")
  eng = DefaultBlockEngine(
    log=l, name="default", 
    verbosity=2,
  )
  with open("xperimental/eth/addrs.txt", "rt") as fd:
    lines = fd.readlines()
    addresses = [line.strip() for line in lines]
    
  dest = addresses[0]
  
  l.P(f"ETH and $R1 transfer test on Ratio1:{eng.evm_network}")
  
  l.P("ETH Test")
  l.P(f"Src  {eng.eth_address} has {eng.web3_get_balance_eth():.4f} ETH")
  l.P(f"Dest {dest} has {eng.web3_get_balance_eth(dest):.4f} ETH")
  l.P(f"Sending 0.1 ETH to {dest}", color='b')
  tx_hash = eng.web3_send_eth(dest, 0.1, wait_for_tx=True, return_receipt=False)
  l.P(f"Executed tx: {tx_hash}", color='g')
  l.P(f"Src  {eng.eth_address} has {eng.web3_get_balance_eth():.4f} ETH")
  l.P(f"Dest {dest} has {eng.web3_get_balance_eth(dest):.4f} ETH")

  l.P("R1 Test")
  l.P(f"Src  {eng.eth_address} has {eng.web3_get_balance_r1():.4f} $R1")
  l.P(f"Dest {dest} has {eng.web3_get_balance_r1(dest):.4f} $R1")
  l.P(f"Sending 100 $R1 to {dest}", color='b')
  tx_hash = eng.web3_send_r1(dest, 100, wait_for_tx=True, return_receipt=False)
  l.P(f"Executed tx: {tx_hash}", color='g')
  l.P(f"Src  {eng.eth_address} has {eng.web3_get_balance_r1():.4f} $R1")
  l.P(f"Dest {dest} has {eng.web3_get_balance_r1(dest):.4f} $R1")
  